import _dialogs
import gtk

gdk = gtk.gdk

TARGET_URILIST = 0
TARGET_RAW = 1

def extract_uris(data):
	"""Convert a text/uri-list to a python list of (still escaped) URIs"""
	lines = data.split('\r\n')
	out = []
	for l in lines:
		if l == chr(0):
			continue	# (gmc adds a '\0' line)
		if l and l[0] != '#':
			out.append(l)
	return out

def provides(context, type): return type in map(str, context.targets)

class RemoteFiles(Exception):
	"Internal use"
	def __init__(self):
		Exception.__init__(self, ('Cannot load files from a remote machine '
			   '(multiple files, or target application/octet-stream not provided)'))

class XDSLoader:
	"""A mix-in class for widgets that can have files/data dropped on
	them. If object is also a GtkWidget, xds_proxy_for(self) is called
	automatically."""

	def __init__(self, types):
		"""Call this after initialising the widget.
		Types is a list of MIME-types, or None to only accept files."""

		targets = [('text/uri-list', 0, TARGET_URILIST)]
		if types:
			for mimetype in types + ['application/octet-stream']:
				targets.append((mimetype, 0, TARGET_RAW))
				
		self.targets = targets
		if isinstance(self, gtk.Widget):
			self.xds_proxy_for(self)
	
	def xds_proxy_for(self, widget):
		"Handle drops on this widget as if they were to 'self'."
		# (Konqueror requires ACTION_MOVE)
		widget.drag_dest_set(gtk.DEST_DEFAULT_MOTION | gtk.DEST_DEFAULT_HIGHLIGHT,
				self.targets,
				gdk.ACTION_COPY | gdk.ACTION_MOVE | gdk.ACTION_PRIVATE)
		
		widget.connect('drag-data-received', self.xds_data_received)
		widget.connect('drag-drop', self.xds_drag_drop)
	
	def xds_drag_drop(self, widget, context, data, info, time):
		"""Called when something is dropped on us. Decide which of the
		offered targets to request and ask for it. xds_data_received will
		be called when it finally arrives."""
		target = widget.drag_dest_find_target(context, self.targets)
		if target is None:
			# Error?
			context.drop_finish(False, time)
		else:	
			widget.drag_get_data(context, target, time)
		return True

	def xds_data_received(self, widget, context, x, y, selection, info, time):
		"Called when we get some data. Internal."
		if selection.data is None:
			# Timeout?
			context.drop_finish(False, time)
			return
		
		if info != TARGET_URILIST:
			return 0

		uris = extract_uris(selection.data)
		
		if len(uris) > 1:
			_dialogs.alert("\nOne file at a time, please!")
			return 0
		
		if not uris:
			_dialogs.alert("Nothing to load!")
			context.drop_finish(False, time)
			return 1

		try:
			try:
				self.xds_load_uris(uris)
			except RemoteFiles:
				if len(uris) != 1 or not provides(context, 'application/octet-stream'):
					raise
				widget.drag_get_data(context, 'application/octet-stream', time)
				return 1	# Don't do drag_finish
		except:
			context.drop_finish(False, time)
			_dialogs.alert("file dropping error")
		else:
			context.drop_finish(True, time)

		return 1
	
	def xds_load_uris(self, uris):
		"""Try to load each URI in the list. Override this if you can handle URIs
		directly. The default method passes each local path to xds_load_from_file()
		and displays an error for anything else.
		The uris are escaped, so a space will appear as '%20'"""
		paths = []
		for uri in uris:
			path = get_local_path(uri)
			if path:
				paths.append(path)
		if len(paths) < len(uris):
			raise RemoteFiles
		for path in paths:
			self.xds_load_from_file(path)
	
	def xds_load_from_file(self, path):
		"""Try to load this local file. Override this if you have a better way
		to load files. The default method opens the file and calls xds_load_from_stream()."""
		try:
			self.xds_load_from_stream(path, None, open(path, 'rb'))
		except:
			_dialogs.info("\nFile type does not seem to be supported")
	

################## #########################
_host_name = None
def our_host_name():
	"""Try to return the canonical name for this computer. This is used
	in the drag-and-drop protocol to work out whether a drop is coming from
	a remote machine (and therefore has to be fetched differently)."""
	from socket import getfqdn
	global _host_name
	if _host_name:
		return _host_name
	try:
		_host_name = getfqdn()
	except:
		_host_name = 'localhost'
		_dialogs.alert("Finding host via socket.getfqdn() failed!")
	return _host_name

def escape(uri):
	"Convert each space to %20, etc"
	import re
	return re.sub('[^:-_./a-zA-Z0-9]',
		lambda match: '%%%02x' % ord(match.group(0)),
		_to_utf8(uri)[0])

def unescape(uri):
	"Convert each %20 to a space, etc"
	if '%' not in uri: return uri
	import re
	return re.sub('%[0-9a-fA-F][0-9a-fA-F]',
		lambda match: chr(int(match.group(0)[1:], 16)),
		uri)

def get_local_path(uri):
	"""Convert 'uri' to a local path and return, if possible. If 'uri'
	is a resource on a remote machine, return None. URI is in the escaped form
	(%20 for space)."""
	if not uri:
		return None

	if uri[0] == '/':
		if uri[1:2] != '/':
			return unescape(uri)	# A normal Unix pathname
		i = uri.find('/', 2)
		if i == -1:
			return None	# //something
		if i == 2:
			return unescape(uri[2:])	# ///path
		remote_host = uri[2:i]
		if remote_host == our_host_name():
			return unescape(uri[i:])	# //localhost/path
		# //otherhost/path
	elif uri[:5].lower() == 'file:':
		if uri[5:6] == '/':
			return get_local_path(uri[5:])
	elif uri[:2] == './' or uri[:3] == '../':
		return unescape(uri)
	return None
